"""
    Module that implements the index structure to storage and process collections.
"""

import pickle
from os import path
from collections import defaultdict
from typing import List, Set, DefaultDict

from matchup.structure.occurrence import Occurrence
from matchup.structure.index.base_index import Index

from matchup.presentation.sanitizer import Sanitizer
from matchup.presentation.text import Term
from matchup.presentation.formats import ExtensionNotSupported, get_file


class InvertedIndex(Index):
    """
        Simple structure based in a inverted file.
    """
    def __init__(self):
        self._inverted_file = defaultdict(list)

    def __str__(self) -> str:
        """
            Transform an index object in an String
        :return: String that represents the structure of index
        """
        index = ""
        for key in self._inverted_file.keys():
            index += str(key + ":" + str(self._inverted_file[key]) + "\n")
        return index

    def __contains__(self, item: str) -> bool:
        """
            This function enables the user to make the associative operation with 'in'
        :param item: keyword
        :return: boolean flag that return true if the keyword are in index
        """
        return item in self._inverted_file

    def __getitem__(self, item: str) -> List[Occurrence]:
        """
            Get some occurrences by item, or init it on data structure
        :param item: keyword
        :return: Occurrences of keyword
        """
        return self._inverted_file[item]

    def load(self, **kwargs) -> Set[str]:
        """
            This is a function that recover the index previously generated.
            Needs the keyword param 'path' that indicates the path to load the index.
        :return: set of files retrieved
        """
        file_path = kwargs.get("path")
        if file_path and path.exists(file_path):
            self._inverted_file.clear()
            with open(file_path, mode='rb') as file:
                self._inverted_file = pickle.load(file)
                return self.__retrieve_file_names()
        raise FileNotFoundError

    def save(self, **kwargs) -> bool:
        """
            Persist data structure on disc.
            Needs the keyword param 'path' that indicates the path to save the index.
        :return: boolean flag that indicates if the data structure can be persisted.
        """
        file_path = kwargs.get("path")
        if file_path and self._inverted_file:
            self.__sort()
            with open(file_path, mode='wb') as file:
                pickle.dump(self._inverted_file, file)
            return True
        raise ReferenceError("You should to process some collection files")

    @property
    def keys(self) -> list:
        """
            Get all keywords presents in index
        :return: list of all keywords
        """
        return list(self._inverted_file.keys())

    def documents_with_keywords(self, kwds: Set[str]) -> Set[str]:
        answer = list()
        for key in self._inverted_file:
            if key in kwds:
                answer += [oc.doc() for oc in self._inverted_file[key]]

        incomplete = {elem for elem in answer if answer.count(elem) != len(kwds)}

        return set(answer) - incomplete

    def maximum_frequencies_per_document(self) -> DefaultDict[str, float]:
        """
            Return one dictionary with structure : Document -> Maximum frequency of one term in it document.
        :return:
        """
        map_docs = defaultdict(float)
        for key in self._inverted_file:
            for occurrence in self._inverted_file[key]:
                if map_docs[occurrence.doc()] < occurrence.frequency:
                    map_docs[occurrence.doc()] = occurrence.frequency
        return map_docs

    def process(self, files: Set[str], **kwargs) -> None:
        """
            This function try to process all content of files that have been inserted before, generating
            the index data structure ready for use.
            Needs the param 'sanitizer' that indicates the sanitizer object to clean the collection.
        :return: None
        """
        sanitizer = kwargs.get("sanitizer")
        sanitizer = sanitizer if sanitizer else Sanitizer()

        for file_name in files:
            try:
                file = get_file(file_name)
            except ExtensionNotSupported:
                continue
            try:
                text_io = file.open()
                self.__process_file(file_name, text_io, sanitizer)
            except ExtensionNotSupported:
                continue
            finally:
                file.close()

    def __process_file(self, filename, content_file, sanitizer) -> None:
        """
            Process one file line by line
        :param filename: name of file
        :param content_file: text file content
        :param sanitizer: sanitizer object
        :return:
        """
        number_line = 1
        for content_line in content_file:
            terms = sanitizer.sanitize_line(content_line, number_line)
            self.__push(terms, filename)
            number_line += 1

    def __push(self, terms: List[Term], file_name: str) -> None:
        """
            Function that push all file terms in index
        :param terms: list with all file terms
        :param file_name: path of file
        :return: None
        """
        for term in terms:
            try:
                idx = self._inverted_file[term.word].index(file_name)
                occurrence = self._inverted_file[term.word][idx]
                occurrence.add(position=term.position)

            except ValueError:
                occurrence = Occurrence(file_name, term)
                self._inverted_file[term.word].append(occurrence)

    def __retrieve_file_names(self) -> Set[str]:
        """
            Iterate the data structure retrieving the file names that generated it index
        :return: set of files retrieved
        """
        files = set()
        for keyword in self._inverted_file:
            for occurrence in self._inverted_file[keyword]:
                files.add(occurrence.doc())
        return files

    def __sort(self) -> None:
        """
            Order the documents in index structure.
        :return: None
        """
        for key in self._inverted_file:
            self._inverted_file[key] = sorted(self._inverted_file[key], key=Occurrence.doc)



"""
    Module that represents the search solution using some IR model in addition to some operations for presentation
"""

from typing import List

from collections import namedtuple

Result = namedtuple("Result", "document score")


class Solution:
    """
        The solution class has the function of properly storing and displaying the responses of the queries
    """
    def __init__(self, results: List[Result]):
        """
            Solution constructor
        :param results: List of (Document, Score).
        """
        self._results = results
        self._idf = None
        self._tf = None

    def __repr__(self) -> str:
        """
            Ranking string
        :return:
        """
        string = ""
        if self._results:
            for terms in self._results:
                doc = terms.document.split('/')[-1]
                string += f"\n{doc} : {terms.score}"
        else:
            string += "No results found."
        return string

    @property
    def results(self) -> List[Result]:
        """
            Ranking List
        :return: List of results
        """
        return self._results

    def __contains__(self, item: Result):
        """
            Boolean flag that indicates if some item are in result list
        :param item:
        :return:
        """
        return item in self._results

    def str_n(self, n) -> str:
        """
            Take the ranking for the first N results.
        :param n:
        :return:
        """
        return str(Solution(self._results[:n]))



"""
    Module that has two important structures used throughout the library: Term and Line
"""

from collections import namedtuple

Term = namedtuple("Term", "word position")
Line = namedtuple("Line", "content number")



import unittest

from matchup.structure.solution import Result
from matchup.models.algorithms import Boolean

from . import set_up_txt_test, set_up_pdf_test


class BooleanTest(unittest.TestCase):
    def test_txt_search_known_response(self):
        self._vocabulary, self._query = set_up_txt_test()
        self._query.ask(answer="artilheiro brasil 1994 gols")
        response = self._query.search(model=Boolean())

        some_expected_results = [Result("./tests/static/files/d1.txt", 1.0),
                                 Result("./tests/static/files/d3.txt", 1.0),
                                 Result("./tests/static/files/d15.txt", 0.75),
                                 Result("./tests/static/files/d11.txt", 0.5)]

        for expected in some_expected_results:
            self.assertTrue(expected in response)

    def test_pdf_search_known_response(self):
        self._vocabulary, self._query = set_up_pdf_test()
        self._query.ask(answer="artilheiro brasil 1994 gols")
        response = self._query.search(model=Boolean())

        some_expected_results = [Result("./tests/static/pdf-files/d1.pdf", 1.0),
                                 Result("./tests/static/pdf-files/d3.pdf", 1.0),
                                 Result("./tests/static/pdf-files/d15.pdf", 0.75),
                                 Result("./tests/static/pdf-files/d11.pdf", 0.5)]

        for expected in some_expected_results:
            self.assertTrue(expected in response)



"""
    Package responsible to get transparent all files independent of its extensions.
"""

__all__ = ['file', 'tools', 'get_file', 'SUPPORTED_EXTENSIONS', 'ExtensionNotSupported']


from matchup.presentation.formats.file import Txt, Pdf, get_extension


SUPPORTED_EXTENSIONS = ['', '.txt', '.pdf']


class ExtensionNotSupported(RuntimeError):
    ...


def get_file(file_path: str):
    extension = get_extension(file_path)

    if extension not in SUPPORTED_EXTENSIONS:
        raise ExtensionNotSupported(f"{extension} is not supported.")

    if extension == '.txt' or not extension:
        return Txt(file_path)
    elif extension == '.pdf':
        return Pdf(file_path)



"""
    Belief Network module.
    Defines a IR modern model network-based. Classic probabilistic and vector space concepts are important here.
"""

from math import sqrt
from typing import List, DefaultDict
from collections import defaultdict

from matchup.models.model import IterModel, Term, Result, Vocabulary


class BeliefNetwork(IterModel):

    def __init__(self):
        super(BeliefNetwork, self).__init__()

    def run(self, query, vocabulary: Vocabulary) -> List[Result]:
        """
            Belief Network algorithm.
        :param query: list of all query terms
        :param vocabulary: data structure that represents the vocabulary
        :return: list of solution -> (document, score)
        """

        query_terms = query.search_input

        self.initialize(query_terms, vocabulary)  # initialize query and vocabulary pointers

        query_repr = self.query_repr(query, vocabulary.idf, vocabulary.tf)  # generate first query abstraction

        self.process_vocabulary_query_based(query_terms, vocabulary)  # process vocabulary based in query keywords

        scores = defaultdict(float)

        t = len(self._term_occurrences) if len(self._term_occurrences) <= 10 else 10

        prob_k = (1/2) ** t  # prob_k : belief_network model.

        while not self.stop():
            doc, doc_repr = self.iter()   # pointer-based document choose
            scores[doc] = self.generate_scores(doc_repr, query_repr, prob_k)  # generate score : belief_network

        scores = sorted(scores.items(), key=lambda v: v[1], reverse=True)
        return self.cast_solution(scores)

    @classmethod
    def generate_scores(cls, doc_repr: DefaultDict[str, float], query_repr: DefaultDict[str, float], prob_k: float)\
            -> float:
        """
            Calculate similarity between one document and query based in Belief Network IR model.
        :param doc_repr: Document vector space representation
        :param query_repr: Query vector space representation
        :param prob_k: Probability that a random key 'k' is relevant
        :return Relevant score of document based in query.
        """
        wij_square_sum = 0.0
        for key in doc_repr.keys():
            wij_square_sum += doc_repr[key] ** 2
        wij_square_sum = sqrt(wij_square_sum)

        wiq_square_sum = 0.0
        for key in query_repr.keys():
            wiq_square_sum += query_repr[key] ** 2
        wiq_square_sum = sqrt(wiq_square_sum)

        score = 0.0
        for key in doc_repr.keys():
            prob_query = query_repr[key] / wiq_square_sum if wiq_square_sum != 0.0 else 0.0
            prob_document = doc_repr[key] / wij_square_sum if wij_square_sum != 0.0 else 0.0
            score += prob_document * prob_query * prob_k

        return score




import unittest

from matchup.presentation.text import Term
from matchup.structure.query import Query
from matchup.structure.vocabulary import Vocabulary


class QueryTest(unittest.TestCase):
    def setUp(self):
        vocabulary = Vocabulary("./tests/static/files", stopwords="./tests/static/pt-br.txt")
        vocabulary.import_collection()

        self._query = Query(vocabulary=vocabulary)

    def test_ask(self):
        answer = 'eu sou o marcos\ntop de linha'
        processed = [Term("marcos", "1-9"), Term("top", "2-0"), Term("linha", "2-7")]

        self._query.ask(answer=answer)
        self.assertEqual(self._query.search_input, processed)



"""
    Module that represents one weighting param for IR models: Term Frequency (TF).
"""


from abc import ABC
from math import log
from collections import defaultdict

TF_ALGORITHMS = {"Binary", "TermFrequency", "LogNormalization", "DoubleNormalization", "DoubleNormalizationK"}


class TFFactory:
    """
        Factory for TF based on String values.
    """
    @staticmethod
    def create_tf_by_str(value: str) -> "TF":
        """
            Evaluate one string and return the correspondent TF.
        :param value:
        :return:
        """
        if value in TF_ALGORITHMS:
            return eval(value)


class TF(ABC):
    """
        Abstract base class who represents TF param.
    """

    def __init__(self, **kwargs):
        """
            Initialize TF structure.
        :param kwargs:
        """
        self._tfs = defaultdict(float)

    def __repr__(self):
        string = ""
        for key in self._tfs:
            string += f"{key} : {self._tfs[key]}\n"
        return string

    def take(self, *,  value: int = 0, reverse: bool = True):
        if value != 0:
            return sorted(self._tfs.items(), key=lambda v: v[1], reverse=reverse)[:value]
        return sorted(self._tfs.items(), key=lambda v: v[1], reverse=reverse)

    def calculate(self, keyword: str, occurrence, document_maximum_frequency, persist: bool = True) -> float:
        """
           Generate TF based in TFType
        :param keyword: keyword to calculate tf
        :param occurrence: keyword occurrence in it document
        :param document_maximum_frequency: maximum frequency in its document
        :param persist: persists in-memory or not the tf score generated.
        :return:float tf
        """
        ...

    def _fij(self, occurrence, document_maximum_frequency):
        return occurrence.frequency / document_maximum_frequency

    def _persist(self, key, tf) -> float:
        self._tfs[key] = tf
        return tf


class Binary(TF):

    def __init__(self, **kwargs):
        super(Binary, self).__init__(**kwargs)

    def calculate(self, keyword: str, occurrence, document_maximum_frequency, persist: bool = True):
        """
            Model to calculate TF based in binary values (0,1)
        :return: float tf
        """
        fij = self._fij(occurrence, document_maximum_frequency)

        if fij > 0:
            return self._persist((keyword, occurrence.doc()), 1.0) if persist else 1.0
        else:
            return self._persist((keyword, occurrence.doc()), 0.0) if persist else 0.0


class TermFrequency(TF):

    def __init__(self, **kwargs):
        super(TermFrequency, self).__init__(**kwargs)

    def calculate(self, keyword: str, occurrence, document_maximum_frequency, persist: bool = True) -> float:
        """
            Model to calculate TF based in fij
        :return: float fij
        """
        fij = self._fij(occurrence, document_maximum_frequency)
        return self._persist((keyword, occurrence.doc()), float(fij)) if persist else float(fij)


class LogNormalization(TF):

    def __init__(self, **kwargs):
        super(LogNormalization, self).__init__(**kwargs)

    def _fij(self, occurrence, document_maximum_frequency):
        return occurrence.frequency

    def calculate(self, keyword: str, occurrence, document_maximum_frequency, persist: bool = True) -> float:
        """
            Model to calculate TF based in : 1 + log fij
        :return: float fij
        """

        fij = self._fij(occurrence, document_maximum_frequency)
        tf = 1 + log(fij, 2)

        return self._persist((keyword, occurrence.doc()), tf) if persist else tf


class DoubleNormalization(TF):
    def __init__(self, **kwargs):
        super(DoubleNormalization, self).__init__(**kwargs)

    def calculate(self, keyword: str, occurrence, document_maximum_frequency, persist: bool = True) -> float:
        """
            Model to calculate TF based in fij
        :return: float fij
        """
        fij = self._fij(occurrence, document_maximum_frequency)

        tf = 0.5 + 0.5 * (fij / document_maximum_frequency)

        return self._persist((keyword, occurrence.doc()), tf) if persist else tf


class DoubleNormalizationK(TF):
    def __init__(self, **kwargs):
        super(DoubleNormalizationK, self).__init__(**kwargs)
        self._K = kwargs.get("K")

    def calculate(self, keyword: str, occurrence, document_maximum_frequency, persist: bool = True) -> float:
        """
            Model to calculate TF based in fij
        :return: float fij
        """
        fij = self._fij(occurrence, document_maximum_frequency)

        K = self._K if self._K is not None else 0.5
        tf = K + (1-K) * (fij / document_maximum_frequency)

        return self._persist((keyword, occurrence.doc()), tf) if persist else tf



"""
    Classic IR model. Probabilistic model are implemented here, based in Interface Model
"""
from collections import defaultdict
from typing import List
from math import log

from matchup.models.model import IterModel, Term, Vocabulary
from matchup.structure.solution import Result


class Probabilistic(IterModel):

    RANGE = 5

    def __init__(self):
        super(Probabilistic, self).__init__()
        self._V = list()
    
    def run(self, query, vocabulary: Vocabulary):
        """
            Principal method that represents IR probabilistic model.
        :param query: list of all query terms
        :param vocabulary: data structure that represents the vocabulary
        :return:
        """
        query_terms = query.search_input

        self.initialize(query_terms, vocabulary)

        self.process_vocabulary_query_based(query_terms, vocabulary)

        rank = self.probabilistic_iterative_perform(vocabulary)

        return rank

    def probabilistic_iterative_perform(self, vocabulary) -> List[Result]:
        """
            Perform the algorithm iterations.
        :param vocabulary:
        :return:
        """
        while True:
            rank = self.iter_rank(vocabulary)
            n_rank = self.__take(rank, self.RANGE)

            if self._V == n_rank:
                return rank

            self.initialize_pointers()  # restart pointers
            self._V = n_rank.copy()

    def iter_rank(self, vocabulary) -> List[Result]:
        """
            One iteration of Probabilistic model execute.
        :return: ranked list of documents, scores
        """
        term_scores = self.process_terms(vocabulary)

        scores = defaultdict(float)
        while not self.stop():
            doc = self.next_doc()

            scores[doc] = self.calculate(doc, term_scores)

        scores = sorted(scores.items(), key=lambda v: v[1], reverse=True)

        return self.cast_solution(scores)

    def process_terms(self, vocabulary):
        """
            Generate scores for all mapped terms.
        :param vocabulary:
        :return:
        """
        term_scores = defaultdict(float)
        for key in self._term_occurrences:
            term_scores[key] = self.score(key, vocabulary)
        return term_scores

    def calculate(self, doc: str, term_scores) -> float:
        """
            Sum the scores of the mapped representation.
        :param doc:
        :param term_scores:
        :return:
        """
        score = 0.0
        for key in self._term_occurrences.keys():
            try:
                occ = self._term_occurrences[key][self._pointers[key]]
                if occ.doc() == doc:
                    score += term_scores[key]
                    self._pointers[key] += 1
            except ValueError:
                continue
            except IndexError:
                continue
        return score

    def score(self, key: str, vocabulary: Vocabulary) -> float:
        """
            Apply probabilistic concepts to calculate the score of one keyword in vocabulary.
        :param key: keyword to generate score
        :param vocabulary: base collection
        :return: float score
        """
        ni = len(vocabulary[key])
        n = len(vocabulary.file_names)

        if self._V:
            vi = self.number_docs_with_key(vocabulary[key])
            v = len(self._V)

            prob_ki_given_r = (vi + (ni / n)) / (v + 1)
            prob_ki_given_not_r = (ni - vi + (ni / v)) / (n - v + 1)
            
        else:
            prob_ki_given_r = 0.5
            prob_ki_given_not_r = ni / n
            
        try:
            score = log(prob_ki_given_r / (1 - prob_ki_given_r), 10) + log((1 - prob_ki_given_not_r)
                                                                           / prob_ki_given_not_r, 10)
        except:
            score = 0.0

        return score

    def number_docs_with_key(self, occurrences) -> int:
        """
            Return the vi_value : number of documents with key
        :param occurrences: Occurrences
        :return:
        """
        doc = 0
        for oc in occurrences:
            if oc.doc() in [t.document for t in self._V]:
                doc += 1
        return doc

    @classmethod
    def __take(cls, rank, n) -> List[Result]:
        """
            Take the first n elements of this list copying to another structure
        :param rank: list of document, score
        :param n: number of elements to take
        :return: sliced list
        """
        cont = 0
        copy = list()
        for i in rank:
            if cont == n:
                return copy
            cont += 1
            copy.append(i)
        return copy



"""
    First abstraction that represents IR models
        Model, IterModel
"""

import abc

from typing import List, Tuple, DefaultDict
from collections import defaultdict

from matchup.structure.solution import Result
from matchup.structure.vocabulary import Vocabulary
from matchup.structure.occurrence import Occurrence

from matchup.presentation.text import Term


class NoSuchModelException(RuntimeError):
    pass


class ModelExecutionError(RuntimeError):
    pass


class Model(abc.ABC):
    """
        IR Models base class.
    """

    @abc.abstractmethod
    def run(self, query, vocabulary: Vocabulary) -> List[Result]:
        """
            Define the principal method of IR models.
        :param query: List of all entry terms
        :param vocabulary: Vocabulary pre-processed
        :return:
        """
        ...

    @classmethod
    def cast_solution(cls, structure: List[tuple]) -> List[Result]:
        return [Result(item[0], round(item[1], 3)) for item in structure]

    @classmethod
    def process_vocabulary_query_based(cls, query: List[Term], vocabulary: Vocabulary) \
            -> DefaultDict[str, List[Occurrence]]:
        """
            Generate document scores based in query
        :param query: query representation
        :param vocabulary: vocabulary structure
        :return: List of occurrences
        """
        target = defaultdict(list)

        idf = vocabulary.idf
        tf = vocabulary.tf

        maximum_frequencies_per_document = vocabulary.maximum_frequencies_per_document()

        for key in query:
            if key.word in vocabulary:
                occurrences = vocabulary[key.word]
                for occurrence in occurrences:
                    score = idf[key.word] * \
                            tf.calculate(key.word, occurrence, maximum_frequencies_per_document[occurrence.doc()])
                    occurrence.score = score
                target[key.word] = occurrences
        return target


class IterModel(Model):
    """
        Describe one variation of Model classes : IterModel classes have some features for help his works
        Pointers and occurrences are implemented here.
    """

    def __init__(self):
        super(IterModel, self).__init__()
        self._term_occurrences = defaultdict(list)
        self._pointers = defaultdict(int)

    @abc.abstractmethod
    def run(self, query, vocabulary: Vocabulary) -> List[Result]:
        """
            Define the principal method of IR models.
        :param query: List of all entry terms
        :param vocabulary: Vocabulary pre-processed
        :return:
        """
        ...

    def initialize(self, query: List[Term], vocabulary: Vocabulary):
        """
            Initialize query-based
        """
        self.initialize_occurrences(query, vocabulary)
        self.initialize_pointers()

    def iter(self) -> Tuple[str, DefaultDict[str, float]]:
        """
            Define one iteration of this iter model algorithm
        :return: doc, doc_repr (keyword -> score)
        """
        doc = self.next_doc()
        doc_repr = self.doc_repr(doc)
        return doc, doc_repr

    def doc_repr(self, doc: str) -> DefaultDict[str, float]:
        """
            Process doc generating it representation
        :param doc: Str represents the lowest document
        :return: doc repr. Dictionary with all term scores by doc. That is the document vector.
        """
        doc_repr = defaultdict(float)
        for key in self._term_occurrences.keys():
            try:
                occ = self._term_occurrences[key][self._pointers[key]]
                if occ.doc() == doc:
                    self._pointers[key] += 1
                    doc_repr[key] = occ.score
            except ValueError:
                continue
            except IndexError:
                continue
        return doc_repr

    def initialize_occurrences(self, query: List[Term], vocabulary: Vocabulary) -> None:
        """
            Create another data structure _term_occurrences that represents the vocabulary with just query
            keywords.
        :param query: original query
        :param vocabulary: original vocabulary
        :return: None
        """
        for key in query:
            if key.word in vocabulary:
                self._term_occurrences[key.word] = vocabulary[key.word]
                
    def initialize_pointers(self) -> None:
        """
            Initialize pointers to model algorithm
        :return: None
        """
        for query_term in self._term_occurrences.keys():
            self._pointers[query_term] = 0

    def next_doc(self) -> str:
        """
            Return the lowest doc pointer by pointers
        :return: lowest document
        """
        lowest = None
        for key in self._term_occurrences.keys():
            try:
                doc = self._term_occurrences[key][self._pointers[key]].doc()
                if lowest and doc < lowest:
                    lowest = doc
                elif not lowest:
                    lowest = doc
            except ValueError:
                continue
            except IndexError:
                continue
        return lowest

    def stop(self) -> bool:
        """
            Given a dictionary with all pointers by keyword and another dictionary with all occurrences by keyword,
            this function calculate if the algorithm is over
        :return: boolean flag indicates if algorithm stops
        """
        for key in self._pointers.keys():
            if self._pointers[key] < len(self._term_occurrences[key]):
                return False
        return True

    def process_vocabulary_query_based(self, query: List[Term], vocabulary: Vocabulary) \
            -> DefaultDict[str, List[Occurrence]]:
        """
            Generate document scores based in query
        :param query: query representation
        :param vocabulary: vocabulary structure
        :return: List of occurrences
        """
        self._term_occurrences = super().process_vocabulary_query_based(query, vocabulary)
        return self._term_occurrences

    @classmethod
    def query_repr(cls, query, vocabulary_idf, vocabulary_tf) -> DefaultDict[str, float]:
        """
            Construct query representation
        :param query: list of all terms
        :param vocabulary_idf: structure IDF for vocabulary
        :param vocabulary_tf: structure TF for vocabulary
        :return: query representation
        """
        maximum_frequency = 0
        occurrences = dict()

        query_terms = query.search_input
        query_tf = query.tf if query.tf else vocabulary_tf
        query_idf = query.idf if query.idf else vocabulary_idf

        for term in query_terms:
            if term.word in occurrences:
                occurrences[term.word].add()
            else:
                occurrences[term.word] = Occurrence()
            f = occurrences[term.word].frequency
            maximum_frequency = f if f > maximum_frequency else maximum_frequency

        query_repr = defaultdict(float)
        for key in query_terms:
            query_repr[key.word] = query_idf[key.word] * query_tf.calculate(key.word, occurrences[key.word],
                                                                            maximum_frequency, persist=False)

        print(query_tf.__class__.__name__, query_repr)

        return query_repr



import unittest

from matchup.structure.occurrence import Occurrence
from matchup.presentation.text import Term


class OccurrenceTest(unittest.TestCase):
    def setUp(self):
        self._occurrence = Occurrence("d3.txt", Term("brasil", "1-15"))

    def test_add_position(self):
        self.assertEqual(self._occurrence.frequency, 1)

        self._occurrence.add(position="2-15")

        self.assertEqual(self._occurrence.frequency, 2)




"""
    All IR algorithms in matchup
"""
from matchup.models.algorithms.extended_boolean import ExtendedBoolean
from matchup.models.algorithms.boolean import Boolean
from matchup.models.algorithms.probabilistic import Probabilistic
from matchup.models.algorithms.vector_space import Vector
from matchup.models.algorithms.generalized_vector import GeneralizedVector
from matchup.models.algorithms.belief_network import BeliefNetwork

__all__ = ["Boolean", "ExtendedBoolean", "Vector", "Probabilistic", "GeneralizedVector", "BeliefNetwork"]



import unittest

from tests.unit import model_test, sanitizer_test, occurrence_test, query_test, vocabulary_test


def create_suite(t_lst, t_load):
    t_lst.append(t_load.loadTestsFromTestCase(model_test.ModelTest))
    t_lst.append(t_load.loadTestsFromTestCase(sanitizer_test.SanitizerTest))
    t_lst.append(t_load.loadTestsFromTestCase(occurrence_test.OccurrenceTest))
    t_lst.append(t_load.loadTestsFromTestCase(query_test.QueryTest))
    t_lst.append(t_load.loadTestsFromTestCase(vocabulary_test.VocabularyTest))


test_list = []
test_loader = unittest.TestLoader()

# add test suites
create_suite(test_list, test_loader)

suite = unittest.TestSuite(test_list)
runner = unittest.TextTestRunner()
runner.run(suite)



"""
    IR modern model. Vector space concepts considering term co-relations
"""
from typing import List, DefaultDict, Set, Tuple
from collections import defaultdict, namedtuple
from math import sqrt
from operator import add

from matchup.models.model import IterModel
from matchup.presentation.text import Term
from matchup.structure.solution import Result
from matchup.structure.vocabulary import Vocabulary

Correlation = namedtuple("Correlation", "keyword cir")
Minterm = namedtuple("Minterm", "correlations id")


class GeneralizedVector(IterModel):
    """
        Class that implements the 'run' method of Generalized Vector IR model.
    """
    def __init__(self):
        super(GeneralizedVector, self).__init__()
        self.minterms = list()

    def run(self, query, vocabulary: Vocabulary) -> List[Result]:
        """
            Run generalized vector model.
        :param query: List of terms.
        :param vocabulary: Vocabulary with a collection.
        :return: Query results
        """

        query_terms = query.search_input

        self.initialize(query_terms, vocabulary)
        self.process_vocabulary_query_based(query_terms, vocabulary)

        query_weight = self.query_repr(query, vocabulary.idf, vocabulary.tf)

        actived_minterms = self.__reduce_correlations(self.__map_correlations(query_weight, vocabulary))

        term_repr, base_len = self.term_repr(actived_minterms)

        doc_repr = self.generalized_doc_repr(term_repr, base_len)
        query_repr = self.generalized_query_repr(query_weight, term_repr, base_len)

        scores = defaultdict(float)
        for doc in doc_repr:
            scores[doc] = self.generalized_calculate(doc_repr[doc], query_repr)

        scores = sorted(scores.items(), key=lambda v: v[1], reverse=True)
        return self.cast_solution(scores)

    @classmethod
    def generalized_calculate(cls, doc: List[float], query: List[float]) -> float:
        """
            Calculate the similarity based on cosine of two vectors: doc vector and query vector.
        :param doc: doc vector
        :param query: query vector
        :return: score of doc
        """
        norm_doc = 0
        norm_query = 0

        for i in range(len(doc)):
            norm_doc += doc[i] ** 2
            norm_query += query[i] ** 2

        norm_query = sqrt(norm_query)
        norm_doc = sqrt(norm_doc)

        intern_prod = 0
        for i in range(len(doc)):
            intern_prod += doc[i] * query[i]

        try:
            cos = intern_prod / (norm_query * norm_doc)
        except:
            cos = 0.0

        return cos

    def generalized_doc_repr(self, term_repr: DefaultDict[str, List[float]], base_len: int) \
            -> DefaultDict[str, List[float]]:
        """
            Generate vectors for all documents.
        :param term_repr: vector of all keys
        :param base_len: len of these vectors
        :return: All document representations
        """
        doc_repr = defaultdict(lambda: [0]*base_len)
        for key in self._term_occurrences:
            for oc in self._term_occurrences[key]:
                kv = list(map(lambda x: x * oc.score, term_repr[key]))
                doc_repr[oc.doc()] = list(map(add, doc_repr[oc.doc()], kv))
        return doc_repr

    @classmethod
    def generalized_query_repr(cls, query_weight: DefaultDict[str, float],  term_repr: DefaultDict[str, List[float]],
                               base_len: int) -> List[float]:
        """
            Generate query vector.
        :param query_weight: query weighting based in keys.
        :param term_repr: term vectors
        :param base_len: base len of term vectors
        :return:
        """
        query_repr = [0]*base_len
        for key in query_weight.keys():
            kv = list(map(lambda x: x * query_weight[key], term_repr[key]))
            query_repr = list(map(add, query_repr, kv))
        return query_repr

    @classmethod
    def term_repr(cls, minterms: List[Minterm]) -> Tuple[DefaultDict[str, List[float]], int]:
        """
            Generate vectors for all terms based in minterms.
        :param minterms:
        :return:
        """
        term_repr = defaultdict(lambda: [0.0]*len(minterms))
        normalization = defaultdict(float)

        count = 0
        for minterm in sorted(minterms, key=lambda x: x.id):
            for correlation in minterm.correlations:
                term_repr[correlation.keyword][count] = correlation.cir
                normalization[correlation.keyword] += correlation.cir**2
            count += 1

        for term in term_repr:
            norm = normalization[term] if normalization[term] != 0 else 1.0
            term_repr[term] = [score / sqrt(norm) for score in term_repr[term]]

        return term_repr, len(minterms)

    @classmethod
    def __map_correlations(cls, query_repr: DefaultDict[str, float], vocabulary: Vocabulary) \
            -> DefaultDict[str, Set[Correlation]]:
        """
            For each document, get the set of minterms that includes it based in query keywords.
        :param query_repr:
        :param vocabulary:
        :return:
        """
        doc_correlations = defaultdict(set)
        for kw in query_repr.keys():
            documents = [(occ.doc(), occ.score) for occ in vocabulary[kw]] if kw in vocabulary else []
            for doc in documents:
                doc_correlations[doc[0]].add(Correlation(kw, doc[1]))
        return doc_correlations

    @classmethod
    def __reduce_correlations(cls, mapped_correlations: DefaultDict[str, Set[Correlation]]) -> List[Minterm]:
        """
            Reduce the mapped collections structure
        :param mapped_correlations:
        :return:
        """
        results = list()
        mtid = 0
        for doc in mapped_correlations:
            element = mapped_correlations[doc]  # elem = set(correlation)
            flag = True
            for result in results:
                if cls.__compare_minterm(result[0], element):
                    flag = False
                    new_correlations = cls.__join(result[0], mapped_correlations[doc])
                    results[results.index(result)] = Minterm(new_correlations, result[1])
            if flag:
                results.append(Minterm(element, mtid))
                mtid += 1
        return results

    @classmethod
    def __compare_minterm(cls, first: Set[Correlation], second: Set[Correlation]) -> bool:
        """
            True if two minterms are similar.
        :param first: first Minterm
        :param second: second Minterm
        :return: Flag that indicates the comparison result.
        """
        return {correlation.keyword for correlation in first} == \
                        {correlation.keyword for correlation in second}

    @classmethod
    def __join(cls, actual: Set[Correlation], other: Set[Correlation]) -> Set[Correlation]:
        """
            Join two minterms.
        :param actual: actual minterm
        :param other: other minterm
        :return:  (actual + other) minterm
        """
        r = set()
        for act_cor in actual:
            for oth_cor in other:
                if oth_cor.keyword == act_cor.keyword:
                    r.add(Correlation(act_cor.keyword, act_cor.cir + oth_cor.cir))
        return r



"""
    Classic IR model. Boolean model are implemented here, based in Interface Model
"""

from collections import defaultdict
from typing import DefaultDict

from matchup.models.model import Model, List, Term, Vocabulary
from matchup.structure.solution import Result


class Boolean(Model):

    def __init__(self):
        super(Boolean, self).__init__()

    def run(self, query, vocabulary: Vocabulary) -> List[Result]:
        """
            Principal method that represents IR boolean model.
        :param query: list of all query terms
        :param vocabulary: data structure that represents the vocabulary
        :return: list of solution -> (document, score)
        """

        query_terms = query.search_input

        scores = self.__map_keywords_in_documents(query_terms, vocabulary)

        maximum_points = len(query_terms)

        rank = defaultdict(float)

        for key in scores.keys():
            w_score = sum(scores[key]) / maximum_points
            if w_score != 0.0:
                rank[key] = w_score

        rank = sorted(rank.items(), key=lambda v: v[1], reverse=True)

        return self.cast_solution(rank)

    @classmethod
    def __map_keywords_in_documents(cls, query: List[Term], vocabulary: Vocabulary) -> DefaultDict[str, list]:
        """
            Given the query and the vocabulary, this function calculates the scores of all words in vocabulary.
        :param query: list of all query terms
        :param vocabulary: data structure that represents the vocabulary
        :return: scores based in document
        """
        scores = defaultdict(list)

        for doc in vocabulary.file_names:
            scores[doc] = []

        for term in query:
            if term.word in vocabulary:
                matched_docs = [occurrence.doc() for occurrence in vocabulary[term.word]]
                for one_doc in matched_docs:
                    scores[one_doc].append(1)
        return scores




"""
    Describes the data structure of IR models design.
"""
from os import path, listdir
from typing import List, DefaultDict, Set

from matchup.structure.occurrence import Occurrence
from matchup.structure.weighting.idf import IDF
from matchup.structure.weighting.tf import TF

from matchup.structure.index.inverted_index import InvertedIndex

from matchup.presentation.sanitizer import Sanitizer


LIB_PATH = path.abspath("./static/lib")
SAVED_FILE_NAME = 'collection.bin'


class Vocabulary:
    """
        Crucial data structure that represents and storage all text processing.
    """

    def __init__(self, save, **kwargs):
        """
        :param response: Path to save processed Vocabulary
        :param kwargs: only accepts 'stopwords', with the stopwords file path
        """
        self._index = InvertedIndex() if kwargs.get('indexer') is None else kwargs.get('indexer')

        self._idf = None
        self._tf = None
        self._path = self.__make_prefix(save) + f"/{SAVED_FILE_NAME}"
        self._sanitizer = self.__make_sanitizer(**kwargs)

        self.file_names = set()

    def import_file(self, file_path: str) -> bool:
        """
            Given a file path of a document, this function append this document into some structure, case the path are
            correct. The processing of this file can be started running function index_files()

        :param file_path: string that represents a relative or absolute path of an txt file
        :return: boolean flag that indicates if the file has been identified
        """
        if path.exists(file_path):
            self.file_names.add(file_path)
            return True
        return False

    def import_folder(self, folder_path: str) -> bool:
        """
            Generalization of import_file(). This function receive a folder path and try to append all documents of
            this folder into some structure. he processing of all this file can be started running function
             index_files()

        :param folder_path: string that represents a relative or absolute path of an folder
        :return: boolean flag that indicates if the folder has been identified
        """
        if path.isdir(folder_path):
            list_dir = filter(lambda x: f'{SAVED_FILE_NAME}' not in x, listdir(folder_path))
            for filename in list_dir:
                self.import_file(folder_path + "/" + filename)
            return True
        raise FileNotFoundError

    def generate_idf(self):
        self._idf.generate(self)

    @property
    def idf(self) -> IDF:
        """
            Get the data structure that represents the IDF weighting
        :return: IDF object
        """
        return self._idf

    @idf.setter
    def idf(self, idf: IDF) -> None:
        """
            This function just calculate IDF of all keywords on vocabulary
        :return: None
        """
        self._idf = idf
        self._idf.generate(self)

    @property
    def tf(self) -> TF:
        """
            Get the data structure that represents the TF weighting
        :return: TF object
        """
        return self._tf

    @tf.setter
    def tf(self, tf: TF) -> None:
        """
            Set the data structure that represents the TF weighting
        :param tf: TF Object
        :return: None
        """
        self._tf = tf

    @property
    def sanitizer(self) -> "Sanitizer":
        """
            Sanitizer property getter
        :return:
        """
        return self._sanitizer

    def import_collection(self) -> bool:
        """
            This is a function that recover the vocabulary previously generated.
        :return: boolean flag that indicates success or failure in case the vocabulary has no generated yet.
        """
        files = self._index.load(path=self._path)
        if files:
            self.file_names = files
            return True
        return False

    def index_files(self) -> None:
        """
            This function try to process all content of files that have been inserted before, generating
            the vocabulary data structure ready for use.
        :return: None
        """
        self._index.process(self.file_names, sanitizer=self._sanitizer)

    def save(self) -> bool:
        """
            Persist data structure on disc.
        :return: boolean flag that indicates if the data structure can be persisted.
        """
        return self._index.save(path=self._path)

    def maximum_frequencies_per_document(self) -> DefaultDict[str, float]:
        return self._index.maximum_frequencies_per_document()

    def documents_with_keywords(self, kwds: Set[str]) -> Set[str]:
        return self._index.documents_with_keywords(kwds)

    @property
    def keys(self) -> List[str]:
        """
            Get all keywords presents in vocabulary
        :return: list of all keywords
        """
        return self._index.keys

    def __str__(self) -> str:
        """
            Transform an Vocabulary object in an String
        :return: String that represents the structure of vocabulary
        """
        return str(self._index)

    def __contains__(self, item: str) -> bool:
        """
            This function enables the user to make the associative operation with 'in'
        :param item: keyword
        :return: boolean flag that return true if the keyword are in vocabulary
        """
        return item in self._index

    def __getitem__(self, item: str) -> List[Occurrence]:
        """
            Get some vocabulary occurrences by item, or initialize it on data structure
        :param item: keyword
        :return: Occurrences of keyword
        """
        return self._index[item]

    @classmethod
    def __make_sanitizer(cls, **kwargs) -> Sanitizer:
        """
            Function that allows the Sanitizer initialization
        :param kwargs: expected: stopwords named param with the stopwords text file path
        :return:
        """
        stopwords = kwargs.get("stopwords")
        stemming = kwargs.get("stemming") if kwargs.get("stemming") else False
        if stopwords:
            return Sanitizer(stopwords_path=stopwords, stemming=stemming)
        else:
            return Sanitizer(stemming=stemming)

    @classmethod
    def __make_prefix(cls, save):
        """
            Function that build the path to save the create_collection
        :param save: path to save
        :return:
        """
        return save if save else LIB_PATH



__all__ = ['boolean_test', 'extended_boolean_test', 'vector_test', 'probabilistic_test', 'main']

from matchup.structure.vocabulary import Vocabulary
from matchup.structure.query import Query


def set_up_txt_test():
    vocabulary = Vocabulary("./tests/static/files", stopwords="./tests/static/pt-br.txt")
    # vocabulary.import_collection()
    vocabulary.import_folder("./tests/static/files")
    vocabulary.index_files()
    vocabulary.save()
    query = Query(vocabulary=vocabulary)
    return vocabulary, query


def set_up_pdf_test():
    vocabulary = Vocabulary("./tests/static/pdf-files", stopwords="./tests/static/pt-br.txt")
    # vocabulary.import_collection()
    vocabulary.import_folder("./tests/static/pdf-files")
    vocabulary.index_files()
    vocabulary.save()
    query = Query(vocabulary=vocabulary)
    return vocabulary, query



"""
    Occurrence module is composed only by the Occurrence class.
"""
from matchup.presentation.text import Term


class Occurrence:
    """
        Occurrence is an encapsulation of:
            document name, keyword, frequency of term in document, positions and score
        Occurrence is an data structure used to compose Vocabulary.
    """
    def __init__(self, doc: str = "", term: Term = None):
        """
            Occurrence constructor.
        :param doc: String that represents the document name
        :param term: Term that encapsulate the keyword and its position in document.
        """
        self._doc = doc
        self._keyword = term.word if term is not None else None
        self._score = 0

        self._frequency = 1
        self._positions = [term.position] if term is not None else []

    def __str__(self) -> str:
        """
        :return: String representation of an Occurrence
        """
        return "('{0}', '{1}', {2}, {3})".format(self._doc, self._keyword, self._frequency, self._positions)

    def __repr__(self) -> str:
        """
        :return: Return the String representation
        """
        return str(self)

    def __eq__(self, doc: str) -> bool:
        """
            Overload the equal operator by document
        :param doc: String that represents the document file name
        :return: boolean flag that indicates if the both documents are equal
        """
        return self._doc == doc

    def add(self, *, position: str = None) -> None:
        """
            Push this occurrence position in doc for the Occurrence structure
        :param position: String that represents the position. Ex : '1-19'
        :return: None
        """
        if position is not None:
            self._positions.append(position)
        self._frequency += 1

    def doc(self) -> str:
        """
            Property that return the document name
        :return: document string name
        """
        return self._doc

    @property
    def frequency(self) -> int:
        """
           Property that return the frequency of term i in document j
        :return: document string name
        """
        return self._frequency

    @property
    def score(self) -> float:
        """
            Property that return the score of this Occurrence
        :return: Score of occurrence
        """
        return self._score

    @score.setter
    def score(self, scr: float) -> None:
        """
            Property setter score
        :param scr: float score
        :return: None
        """
        self._score = scr

    @property
    def keyword(self) -> str:
        """
            Keyword property getter
        :return: Keyword name
        """
        return self._keyword

    @keyword.setter
    def keyword(self, kw: str) -> None:
        """
            Keyword property setter
        :param kw: Keyword to update
        :return: None
        """
        self._keyword = kw



import unittest

from matchup.presentation.sanitizer import Sanitizer
from matchup.presentation.text import Line, Term


class SanitizerTest(unittest.TestCase):
    def setUp(self):
        self._sanitizer = Sanitizer(stopwords_path="./tests/static/pt-br.txt")

    def test_import_stopwords(self):
        response = self._sanitizer.import_stopwords()

        self.assertTrue(response)
        self.assertTrue('quando' in response)

    def test_remove_special_chars(self):
        test_case = '^`}!@#!@#!$%$$a%&%¨*&(*&*¨¨%$$)_^{`:?|/'
        test_response = 'a'

        self.assertEqual(self._sanitizer._remove_special_chars_lower(test_case), test_response)

    def test_remove_accents(self):
        test_case = 'áéíóúàèìòùâêîôûÃ'
        test_response = 'aeiouaeiouaeiouA'

        self.assertEqual(self._sanitizer.strip_accents(test_case), test_response)

    def test_index_line(self):
        line = Line("eu sou o marcos e existem stopwords (em) tudo", 1)
        line_list = ['marcos', 'existem', 'stopwords', 'tudo']
        response_terms = [Term('marcos', '1-9'), Term('existem', '1-18'),
                          Term('stopwords', '1-26'), Term('tudo', '1-41')]

        self.assertEqual(self._sanitizer.index_line(line_list, line), response_terms)

    def test_sanitize_line(self):
        line = "eu sòu o mãrcos e existem stôpwords em túdo"
        response_terms = [Term('marcos', '1-9'), Term('existem', '1-18'),
                          Term('stopwords', '1-26'), Term('tudo', '1-39')]
        self.assertEqual(self._sanitizer.sanitize_line(line, 1), response_terms)








"""
    Module that represents the interface of indexer implementations.
"""
from abc import ABC, abstractmethod
from typing import List, Set, DefaultDict

from matchup.structure.occurrence import Occurrence


class Index(ABC):

    @abstractmethod
    def __str__(self) -> str:
        """
            Simple index string representation
        :return:
        """
        pass

    @abstractmethod
    def __contains__(self, item: str):
        """
            This function enables the user to make the associative operation with 'in'
        :param item: One keyword
        :return:
        """
        pass

    @abstractmethod
    def __getitem__(self, item: str) -> List[Occurrence]:
        """
            Get some occurrences by keyword, or init it on data structure
        :param item: keyword
        :return:
        """
        pass

    def process(self, files: Set[str], **kwargs) -> None:
        """
            This function try to process all content of files that have been inserted before, generating
            the index data structure ready for use.
        :return: None
        """
        pass

    def load(self, **kwargs) -> Set[str]:
        """
            This is a function that recover the index previously generated.
        :return: Set of files that generated its index.
        """
        pass

    def save(self, **kwargs) -> bool:
        """
            Persist data structure on disc.
        :return: boolean flag that indicates if the data structure can be persisted.
        """
        pass

    @property
    @abstractmethod
    def keys(self) -> List[str]:
        """
            Get all keywords presents in vocabulary
        :return: list of all keywords
        """
        pass

    @abstractmethod
    def maximum_frequencies_per_document(self) -> DefaultDict[str, float]:
        """
            Return one dictionary with structure : Document -> Maximum frequency of one term in it document.
        :return:
        """
        pass

    @abstractmethod
    def documents_with_keywords(self, kwds: Set[str]) -> Set[str]:
        """
            Return one set of documents that contains the set of keywords passed with param.
        :param kwds: set of keywords
        :return:
        """
        pass







"""
    The query module is responsible for encapsulating everything related to the process of creating a user query.
"""
from typing import List

from matchup.presentation.sanitizer import Sanitizer
from matchup.models.orchestrator import Orchestrator, Model
from matchup.structure.solution import Solution
from matchup.presentation.text import Term


class NoSuchAnswerException(RuntimeError):
    """
        Exception when no such answer (input) are given by user during a search method.
    """
    pass


class Query:
    """
        Represents the Query of the IR service.
        The query is responsible for processing and generating user input to search a previously built create_collection
    """
    def __init__(self, *, vocabulary, **kwargs):
        stp_path = vocabulary.sanitizer.stopwords_path
        stem = vocabulary.sanitizer.is_stemmig()

        self._sanitizer = Sanitizer(stopwords_path=stp_path, stemming=stem)

        self._orq = Orchestrator(vocabulary)
        self._answer = list()

        self._tf = self._idf = None
        self.process_kwargs(vocabulary, **kwargs)

    def ask(self, answer: str = None) -> None:
        """
            Make query since a command line prompt.
        :return: None
        """
        if not answer:
            self._answer = self._sanitizer.sanitize_line(self._io_answer(), 1)
        else:
            self._answer = self._text_answer(answer)

        # self._orq.entry = self._answer

    @property
    def search_input(self) -> List[Term]:
        """
            Input property getter.
        :return:
        """
        return self._answer

    def search(self, *, model: Model = None, idf=None, tf=None) -> Solution:
        """
            Receive an IR model and execute the query based in user answer and the vocabulary.
        :param model: ModelType that represents the IR model
        :param idf: Describe the class  of IDF
        :param tf: Describe the class of TF
        :return: list of solution -> (document, score)
        """
        results = self._orq.search(self, model, idf, tf)
        return Solution(results)

    @classmethod
    def _io_answer(cls) -> str:
        """
            IO operation to get search input.
        :return:
        """
        message = "{0}\n{1: >18}".format(25 * "= ", "Search: ")
        return input(message)

    def _text_answer(self, plain_answer):
        """
            Sanitize operation for plain text user input.
        :param plain_answer:
        :return:
        """
        number_line = 1
        text = plain_answer.split("\n")
        terms = []
        for line in text:
            line.strip()
            if line:
                terms += self._sanitizer.sanitize_line(line, number_line)
                number_line += 1
        return terms

    def process_kwargs(self, vocabulary, **kwargs):
        """ se a ponderação for igual a do vocabulário, não há necessidade de aumentar o processamento."""
        if "tf" in kwargs and kwargs.get("tf").__class__.__name__ != vocabulary.tf.__class__.__name__:
            self._tf = kwargs.get("tf")

        if "idf" in kwargs and kwargs.get("idf").__class__.__name__ != vocabulary.idf.__class__.__name__:
            self._idf = kwargs.get("idf")
            self._idf.generate(vocabulary)

    @property
    def tf(self):
        return self._tf

    @tf.setter
    def tf(self, tf):
        self._tf = tf

    @property
    def idf(self):
        return self._idf

    @idf.setter
    def idf(self, idf):
        self._idf = idf

    def generate_idf(self, vocabulary):
        if self._idf:
            self._idf.generate(vocabulary)



from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.converter import TextConverter
from pdfminer.layout import LAParams
from pdfminer.pdfpage import PDFPage
from io import StringIO


def convert_pdf_to_txt(path) -> str:
    rsrcmgr = PDFResourceManager()
    retstr = StringIO()
    laparams = LAParams()
    device = TextConverter(rsrcmgr, retstr, laparams=laparams)
    fp = open(path, 'rb')
    interpreter = PDFPageInterpreter(rsrcmgr, device)
    password = ""
    maxpages = 0
    caching = True
    pagenos=set()

    for page in PDFPage.get_pages(fp, pagenos, maxpages=maxpages, password=password, caching=caching,
                                  check_extractable=True):
        interpreter.process_page(page)

    text = retstr.getvalue()

    fp.close()
    device.close()
    retstr.close()
    return text



"""
    Module that represents one weighting param for IR models: Inverse Document Frequency (IDF).
"""

from abc import ABC
from math import log
from collections import defaultdict


IDF_ALGORITHMS = {"Unary", "InverseFrequency", "InverseFrequencySmooth", "InverseFrequencyMax",
                  "ProbabilisticInverseFrequency"}


class IDFFactory:
    """
        Factory for IDF based on String values.
    """
    @staticmethod
    def create_idf_by_str(value: str) -> "IDF":
        """
            Evaluate one string and return the correspondent IDF.
        :param value:
        :return:
        """
        if value in IDF_ALGORITHMS:
            return eval(value)


class IDF(ABC):
    """
        Abstract base class who represents IDF parameter.
    """
    def __init__(self, vocabulary=None):
        """
            Just initialize IDF structure.
        """
        self._idfs = defaultdict(float)
        if vocabulary:
            self.generate(vocabulary)

    def __repr__(self) -> str:
        """
            String representation.
        :return: string
        """
        string = ""
        for key in self._idfs:
            string += f"{key} : {self._idfs[key]}\n"
        return string

    def generate(self, vocabulary) -> None:
        """
            Generate the idf for all keywords in a vocabulary
        :param vocabulary: Vocabulary object
        :return: None
        """
        ...

    def take(self, *, value: int = 0, reverse: bool = True):
        """
            Take the first 'value' elements of IDF dict
        :param value:
        :param reverse : asc or desc.
        :return:
        """
        if value != 0:
            return sorted(self._idfs.items(), key=lambda v: v[1], reverse=reverse)[:value]
        return sorted(self._idfs.items(), key=lambda v: v[1], reverse=reverse)

    def __getitem__(self, item: str) -> float:
        """
            Overload operator [] on idf structure
        :param item: keyword (document)
        :return: float idf value
        """
        return self._idfs[item]


class Unary(IDF):

    def __init__(self, vocabulary=None):
        super(Unary, self).__init__(vocabulary)

    def generate(self, vocabulary):
        """
            Model to calculate IDF based in unary weight (1)
        :param vocabulary: structure to generate IDF
        :return: None
        """
        for key in vocabulary.keys:
            self._idfs[key] = 1


class InverseFrequency(IDF):

    def __init__(self, vocabulary=None):
        super(InverseFrequency, self).__init__(vocabulary)

    def generate(self, vocabulary):
        """
             Model to calculate IDF based in inverse frequency : log N / ni
        :param vocabulary:  structure to generate IDF
        :return: None
        """
        for key in vocabulary.keys:
            self._idfs[key] = log(len(vocabulary.file_names) / len(vocabulary[key]), 10)


class InverseFrequencySmooth(IDF):

    def __init__(self, vocabulary=None):
        super(InverseFrequencySmooth, self).__init__(vocabulary)

    def generate(self, vocabulary):
        """
            Model to calculate IDF based in inverse frequency smooth : log 1 + N / ni
        :param vocabulary: structure to generate IDF
        :return: None
        """
        for key in vocabulary.keys:
            self._idfs[key] = \
                log(1 + (len(vocabulary.file_names) / len(vocabulary[key])), 10)


class InverseFrequencyMax(IDF):

    def __init__(self, vocabulary=None):
        super(InverseFrequencyMax, self).__init__(vocabulary)

    def generate(self, vocabulary):
        """
                Model to calculate IDF based in max inverse frequency  : log 1 + max ni / ni
        :param vocabulary:  structure to generate IDF
        :return: None
        """
        all_number_docs_by_keyword = [len(vocabulary[key]) for key in vocabulary.keys]
        max_ni = max(all_number_docs_by_keyword)

        for key in vocabulary.keys:
            self._idfs[key] = \
                log(1 + (max_ni / len(vocabulary[key])), 10)


class ProbabilisticInverseFrequency(IDF):
    def __init__(self, vocabulary=None):
        super(ProbabilisticInverseFrequency, self).__init__(vocabulary)

    def generate(self, vocabulary):
        """
                Model to calculate IDF based in probabilistic inverse frequency  : log N - ni / ni
                This weighting are not so good when N - ni / ni < 1 :: N - ni < ni :: N < 2ni
        :param vocabulary:  structure to generate IDF
        :return: None
        """
        for key in vocabulary.keys:
            n = len(vocabulary.file_names)
            ni = len(vocabulary[key])
            if n >= 2*ni:
                self._idfs[key] = \
                    log((n - ni) / ni, 10)
            else:
                self._idfs[key] = 0.0



import unittest

from tests.functional import boolean_test, extended_boolean_test, probabilistic_test, vector_test,\
    generalized_vector_test, belief_network_test


def create_suite(t_lst, t_load):
    t_lst.append(t_load.loadTestsFromTestCase(boolean_test.BooleanTest))
    t_lst.append(t_load.loadTestsFromTestCase(extended_boolean_test.ExtendedBooleanTest))
    t_lst.append(t_load.loadTestsFromTestCase(probabilistic_test.ProbabilisticTest))
    t_lst.append(t_load.loadTestsFromTestCase(vector_test.VectorTest))
    t_lst.append(t_load.loadTestsFromTestCase(generalized_vector_test.GeneralizedVectorTest))
    t_lst.append(t_load.loadTestsFromTestCase(belief_network_test.BeliefNetworkTest))


test_list = []
test_loader = unittest.TestLoader()

# add test suites
create_suite(test_list, test_loader)

suite = unittest.TestSuite(test_list)
runner = unittest.TextTestRunner()
runner.run(suite)



"""
    The brain of IR algorithms. This module are responsible to execute one model and return the resulted
    scored document list.
"""
from typing import List

from matchup.structure.solution import Result
from matchup.structure.weighting.idf import InverseFrequency
from matchup.structure.weighting.tf import LogNormalization

from matchup.models.model import Model
from matchup.models.algorithms.vector_space import Vector

from matchup.presentation.text import Term


class NoSuchInputException(RuntimeError):
    pass


class ModelMissingParameters(RuntimeError):
    pass


class Orchestrator:

    def __init__(self, vocabulary):
        self._vocabulary = vocabulary

    def search(self, query, model: Model = None, idf=None, tf=None) -> List[Result]:
        """
            Core function. Execute one IR model based in vocabulary and input(query)
        :param query: Query for IR model
        :param model: IR model to execute
        :param idf: IDF class to weighting IR model
        :param tf: TF class to weighting IR model
        :return: list of solution -> (document, score)
        """

        # setting algorithms IDF weighting
        self._configure_weighting(idf, tf)
        model = model if model else Vector()

        if query.search_input:
            return model.run(query, self._vocabulary)
        else:
            raise NoSuchInputException("You should to put some search. Try again!")

    def _configure_weighting(self, idf=None, tf=None):
        if not tf:
            if not self._vocabulary.tf:
                self._vocabulary.tf = LogNormalization()
        else:
            self._vocabulary.tf = tf

        if not idf:
            if not self._vocabulary.idf:
                self._vocabulary.idf = InverseFrequency()
        else:
            self._vocabulary.idf = idf




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="matchup-ir",
    version="0.2.2",
    author="Marcos Pontes",
    author_email="marcos.rezende@aluno.ufop.com",
    description="A IR simple library.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/marcosfpr/match_up_lib",
    packages=setuptools.find_packages(exclude=['tests', 'docs']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



import unittest

from matchup.structure.vocabulary import Vocabulary


class VocabularyTest(unittest.TestCase):
    def setUp(self):
        self._vocabulary = Vocabulary("./tests/static/files", stopwords="./tests/static/pt-br.txt")

    def test_import_file(self):
        file = "./tests/static/files/d1.txt"
        self.assertTrue(self._vocabulary.import_file(file))
        self.assertTrue(file in self._vocabulary.file_names)

    def test_import_folder(self):
        folder = "./tests/static/files"
        self.assertTrue(self._vocabulary.import_folder(folder))

        self.assertTrue(len(self._vocabulary.file_names) == 20)

    def test_import_vocabulary(self):
        self.assertTrue(self._vocabulary.import_collection())
        self.assertTrue(self._vocabulary.keys is not None)
        self.assertTrue('brasil' in self._vocabulary.keys)

    def test_generate_vocabulary(self):
        self.assertTrue(not self._vocabulary.keys)

        folder = "./tests/static/files"
        self._vocabulary.import_folder(folder_path=folder)

        self._vocabulary.index_files()

        self.assertTrue('brasil' in self._vocabulary.keys)

    def test_generate_vocabulary_pdf(self):
        self.assertTrue(not self._vocabulary.keys)

        folder = "./tests/static/pdf-files"
        self._vocabulary.import_folder(folder_path=folder)

        self._vocabulary.index_files()

        self.assertTrue('brasil' in self._vocabulary.keys)



"""
    Classic IR model. Vector model are implemented here, based in Interface Model
"""
from math import sqrt

from collections import defaultdict
from typing import List, DefaultDict

from matchup.models.model import IterModel, Term, Vocabulary
from matchup.structure.weighting.tf import DoubleNormalization
from matchup.structure.weighting.idf import InverseFrequency
from matchup.structure.solution import Result


class Vector(IterModel):

    def __init__(self):
        super(Vector, self).__init__()

    def run(self, query, vocabulary: Vocabulary) -> List[Result]:
        """
           Principal method that represents IR vector space model.
        :param query: list of all query terms
        :param vocabulary: data structure that represents the vocabulary
        :return: list of solution -> (document, score)
        """
        query_terms = query.search_input

        self.initialize(query_terms, vocabulary)

        query_repr = self.query_repr(query, vocabulary.idf, vocabulary.tf)

        self.process_vocabulary_query_based(query_terms, vocabulary)

        scores = defaultdict(float)

        while not self.stop():
            doc, doc_repr = self.iter()
            scores[doc] = self.generate_scores(doc_repr, query_repr)

        scores = sorted(scores.items(), key=lambda v: v[1], reverse=True)
        return self.cast_solution(scores)

    @classmethod
    def generate_scores(cls, doc_repr: DefaultDict[str, float], query_repr: DefaultDict[str, float]) -> float:
        """
            Calculate the similarity between one doc and one query by its representations
        :param doc_repr: doc repr
        :param query_repr: query repr
        :return: similarity
        """
        norm_query = 0
        norm_doc = 0
        intern_product = 0

        for key in query_repr.keys():
            intern_product += (doc_repr[key] * query_repr[key])
            norm_doc += doc_repr[key] ** 2
            norm_query += query_repr[key] ** 2

        norm_query = sqrt(norm_query)
        norm_doc = sqrt(norm_doc)

        if norm_doc and norm_query:
            return intern_product / (norm_doc * norm_query)
        else:
            return 0.0



"""
    IR modern model. Combination with Boolean and Vector Space concepts.
"""

from math import pow

from typing import List
from collections import defaultdict

from matchup.presentation.text import Term
from matchup.models.model import IterModel, Vocabulary


class ExtendedBoolean(IterModel):

    def __init__(self, p: float):
        super(ExtendedBoolean, self).__init__()
        self.p = p

    def run(self, query, vocabulary: Vocabulary):
        """
            Needs weighting <TF <TermFrequency> x IDF <InverseFrequency>>
        :param query:
        :param vocabulary:
        :return:
        """
        query_terms = query.search_input

        self.initialize_occurrences(query_terms, vocabulary)
        self.initialize_pointers()

        self._term_occurrences = self.process_vocabulary_query_based(query_terms, vocabulary)

        scores = defaultdict(float)

        while not self.stop():
            doc, doc_repr = self.iter()
            scores[doc] = self.generate_scores(doc_repr, len({term.word for term in query_terms}))  # group keys

        scores = sorted(scores.items(), key=lambda v: v[1], reverse=True)
        return self.cast_solution(scores)

    def generate_scores(self, doc_repr, m) -> float:
        """
            Generate scores with Extended Boolean Model
        :param doc_repr: vector that represents one scores of terms in one document
        :param m: size of query (group by key)
        :return:
        """
        score = 0.0
        for key in doc_repr:
            score += pow((1 - doc_repr[key]), self.p)
            
        compensation = m - len(doc_repr)
        if compensation:
            score += compensation

        score /= m

        try:
            score = pow(score, 1/self.p)
        except ValueError:
            score = 0.0

        return 1.0 - score









import abc
import os

TEMP_SUFFIX = "_temp.txt"


def get_extension(filename: str) -> str:
    extension = -1
    return os.path.splitext(filename)[extension]


def get_base_name(filename: str) -> str:
    basename = 0
    return os.path.splitext(filename)[basename]


class AbstractFile(abc.ABC):

    @abc.abstractmethod
    def open(self):
        """
        Define the mode to open the file based in its extension.
        :return: text file object
        """
        ...

    @abc.abstractmethod
    def close(self) -> None:
        """
            Just close the file on a security way.
        :return: None
        """
        ...

    @abc.abstractmethod
    def content(self) -> str:
        """
            Return string that represents all text file
        :return:
        """
        ...


class Pdf(AbstractFile):

    def __init__(self, file_path):
        self._file_path = file_path
        self._file = None

    def open(self):
        from .tools import convert_pdf_to_txt
        text = convert_pdf_to_txt(self._file_path)
        self._file = open(get_base_name(self._file_path) + TEMP_SUFFIX, mode='w+', encoding='utf-8', errors='ignore')
        self._file.write(text)
        self._file.seek(0)
        return self._file

    def close(self):
        file_path = self._file.name
        self._file.close()
        os.remove(file_path)

    def content(self) -> str:
        from .tools import convert_pdf_to_txt
        return convert_pdf_to_txt(self._file_path)


class Txt(AbstractFile):

    def __init__(self, file_path):
        self._file_path = file_path
        self._file = None

    def open(self):
        self._file = open(self._file_path, mode='r', encoding='utf-8', errors='ignore')
        return self._file

    def close(self):
        self._file.close()

    def content(self) -> str:
        f = open(self._file_path, mode='r', encoding='utf-8', errors='ignore')
        content = f.read()
        f.close()
        return content



import unittest

from matchup.models.model import Model
from matchup.structure.solution import Result


class ModelTest(unittest.TestCase):

    def test_cast_solution(self):
        ipt = [("d1.txt", 1.0), ("d2.txt", 0.421)]
        opt = [Result("d1.txt", 1.0), Result("d2.txt", 0.421)]

        self.assertEqual(Model.cast_solution(ipt), opt)



"""
    Module responsible to configure the text representation.
"""

import re
import unicodedata
from typing import List, Set

from matchup.presentation.text import Term, Line


class Sanitizer:
    """
      Responsible to clean and process the text representation.
    """

    def __init__(self, *, stopwords_path: str = None, stemming: bool = False):
        """
            Sanitizer constructor
        :param stopwords_path: Path that contains a file with stopwords. This file can't have special characters.
        """
        self._stopwords = set()
        self._stemmer = None

        self._set_stemming(stemming)

        if stopwords_path:
            self._stopwords_path = stopwords_path
            self._stopwords = self.import_stopwords()

    def sanitize_line(self, line: str, number_line: int) -> List[Term]:
        """
            This function sanitize one line. The number is basically for presentation
        :param line: Complete line to be processed
        :param number_line: number of line
        :return:
        """
        base_line = Line(line, number_line)

        line_cleaned = base_line.content.strip()
        line_cleaned = self.strip_accents(line_cleaned)

        words = line_cleaned.split()
        indexed = self.index_line(words, base_line)

        return indexed

    def import_stopwords(self) -> Set[str]:
        """
            Retrieve stopwords from a file.
        :return: set of stopwords
        """
        with open(self._stopwords_path, mode='r', encoding='utf-8') as file:
            self._stopwords = {line.strip() for line in file}
        return self._stopwords

    def add_stopwords(self, stopwords: Set[str]):
        """
            Add a set of stopwords manually.
        :param stopwords: set of stopwords.
        :return:
        """
        self._stopwords = self._stopwords.union(stopwords)

    @property
    def stopwords_path(self) -> str:
        """
            Property that get the stopwords file path
        :return: Complete stopwords file path
        """
        return self._stopwords_path

    @stopwords_path.setter
    def stopwords_path(self, path: str) -> None:
        """
            Property setter stopwords path
        :param path: New stopwords path
        :return: None
        """
        self._stopwords_path = path

    def is_stemmig(self):
        return self._stemmer is not None

    def index_line(self, words: List[str], base_line: Line) -> List[Term]:
        """
            This function index one line and returning all words sanitized
            The list must be sorted by occurrence in text!
        :param words: base-line stripped and without stopwords
        :param base_line: line to be sanitized
        :return: list of indexed words : list(Term)
        """

        filtered = self._filter_stopwords(words)
        base_line_stripped = Sanitizer.strip_accents(base_line.content).lower()

        indexed_words = []

        acc_value = 0
        for word in filtered:

            word_position = base_line_stripped[acc_value::].find(word)
            acc_value += word_position

            position = str(base_line.number) + "-" + str(acc_value)

            indexed_words.append(Term(self._stemmer.stem(word) if self._stemmer else word, position))

            acc_value += len(word)

        return indexed_words

    @staticmethod
    def strip_accents(text: str) -> str:
        """
            Strip accents of one text
        :param text: old text
        :return: text sanitized
        """
        text = unicodedata.normalize('NFD', text).encode('ascii', 'ignore').decode("utf-8")
        return str(text)

    @staticmethod
    def _remove_special_chars_lower(word: str) -> str:
        """
            Given an word, this function clean this string, removing the special chars
        :param word: string to be sanitized
        :return:
        """
        return re.sub('[^a-zA-Z0-9áéíóúàèìòùãõâêîôû\\-\n]', '', word).lower()

    def _filter_stopwords(self, words: List[str]) -> List[str]:
        filtered = list()
        for word in words:
            cleaned_word = self._remove_special_chars_lower(word)
            if cleaned_word and cleaned_word not in self._stopwords:
                filtered.append(cleaned_word)
        return filtered

    def _set_stemming(self, stemming):
        if stemming:
            from nltk.stem import PorterStemmer
            self._stemmer = PorterStemmer()
